import argparse

from workbench.prediction.model_builder import create_model
from workbench.prediction.model_deployer import deploy_model
from workbench.executor import execute_local_notebook

import warnings

def main():
    warnings.filterwarnings("ignore")

    parser = argparse.ArgumentParser(prog="workbench",
                                     description="GCP Vertex AI Workbench CLI")
    parser.add_argument("action", choices=["build", "deploy", "execute-notebook"],
                                help="action")
    parser.add_argument("--tag", type=str, dest="tag", required=False,
                                help="Docker Tag To Use With Build And/Or Deploy")


    execute_notebook_group = parser.add_argument_group("execute-notebook")
    execute_notebook_group.add_argument("--gcs-folder", type=str, dest="gcs_folder", required=True,
                                help="required for storing your notebook as well as the output")
    execute_notebook_group.add_argument("--notebook", type=str, dest="notebook", required=True,
                                help="path for notebook to execute")
    execute_notebook_group.add_argument("--vm-type", type=str, dest="vm_type",
                                help="type of VM to use for execution", default="n1-standard-4")
    execute_notebook_group.add_argument("--env-container", type=str, dest="env_container",
                                help="Docker container, should be basead on GCP Deep Learnign Containers", 
                                default="gcr.io/deeplearning-platform-release/base-cu110:latest")
    execute_notebook_group.add_argument("--kernel", type=str, dest="kernel",
                                help="kernels to use, your container should be compatible with this kernel", 
                                default="python3")
    execute_notebook_group.add_argument("--execution-id", type=str, dest="execution_id",
                                help="execution id, should be unique or leave empty", 
                                default="")
    location_argument = execute_notebook_group.add_argument("--location", type=str, dest="location",
                                help="location should be supported by Vertex AI Workbench ")
    project_argument = execute_notebook_group.add_argument("--project", type=str, dest="project",
                                help="GCP Project (Vertex AI API should be enabled)")


    builder_group = parser.add_argument_group("build")
    builder_group.add_argument("--path", type=str, dest="path", default=".",
                                help="path to prediction.py file")
    deployer_group = parser.add_argument_group("deploy")
    deployer_group.add_argument("--name", type=str, dest="name",
                                help="model name")
    deployer_group._group_actions.append(location_argument)
    deployer_group._group_actions.append(project_argument)

    args = parser.parse_args()

    if args.action == "build":
        create_model(args.tag, args.path)
    if args.action == "deploy":
        deploy_model(args.project, args.location, args.name, args.tag)
    if args.action == "execute-notebook":
        result = execute_local_notebook(args.project, 
                                args.location, 
                                args.notebook, 
                                args.gcs_folder, 
                                args.execution_id, 
                                args.env_container, 
                                args.kernel, 
                                args.vm_type)
        if not result:
            print("Execution failed")
        else:
            print("Result will be visible at: " + result["viewer_url"])
            print("Executed notebook will be at: " + result["notebook_gcs_url"])
            print("Execution id: " + result["execution_uri"])
            print("Operation id: " + result["operation_uri"])



from typing import Optional, Dict, Sequence, Tuple

from google.cloud import aiplatform

import yaml


def deploy_model(project: str,
    location: str,
    display_name: str,
    serving_container_image_uri: str,):
    print("Uploading model")
    model = _upload_model(project, location, display_name, serving_container_image_uri)
    print(f"model: {model.resource_name}")
    print("Deploying model (this might take a LONG.... time)")
    _deploy_model(project, location, model.resource_name)


def deploy_model_with_confg(
    display_name: str,
    serving_container_image_uri: str,
    config_path: str,):
    with open(config_path, "r") as f:
        config = yaml.load(f)
    print("Uploading model")
    model = _upload_model(
        config["project"], config["location"], 
        display_name, serving_container_image_uri)
    print(f"model: {model.resource_name}")
    print("Deploying model (this might take a LONG.... time)")
    _deploy_model_with_config(model.resource_name, config_path)



def _upload_model(
    project: str,
    location: str,
    display_name: str,
    serving_container_image_uri: str,
):

    aiplatform.init(project=project, location=location)

    model = aiplatform.Model.upload(
        display_name=display_name,
        serving_container_image_uri=serving_container_image_uri,
        serving_container_predict_route="/predict",
        serving_container_health_route="/ping",
        serving_container_ports=[8080],
        sync=True,
    )

    model.wait()

    print(model.display_name)
    print(model.resource_name)
    return model


def _deploy_model_with_config(model_name: str, config) -> aiplatform.Model:
    config["model_name"] = model_name
    model = _deploy_model(**config)

    return model


def _deploy_model(
    project,
    location,
    model_name: str,
    endpoint: Optional[aiplatform.Endpoint] = None,
    deployed_model_display_name: Optional[str] = None,
    traffic_percentage: Optional[int] = 100,
    traffic_split: Optional[Dict[str, int]] = None,
    machine_type: Optional[str] = "n1-standard-2",
    min_replica_count: int = 1,
    max_replica_count: int = 1,
    accelerator_type: Optional[str] = None,
    accelerator_count: Optional[int] = None,
    service_account: Optional[str] = None,
    explanation_metadata = None,
    explanation_parameters = None,
    metadata: Optional[Sequence[Tuple[str, str]]] = (),
    encryption_spec_key_name: Optional[str] = None,
):
    aiplatform.init(project=project, location=location)

    model = aiplatform.Model(model_name=model_name)

    model.deploy(
        endpoint=endpoint,
        deployed_model_display_name=deployed_model_display_name,
        traffic_percentage=traffic_percentage,
        traffic_split=traffic_split,
        machine_type=machine_type,
        min_replica_count=min_replica_count,
        max_replica_count=max_replica_count,
        accelerator_type=accelerator_type,
        accelerator_count=accelerator_count,
        service_account=service_account,
        explanation_metadata=explanation_metadata,
        explanation_parameters=explanation_parameters,
        metadata=metadata,
        encryption_spec_key_name=encryption_spec_key_name,
        sync=True,
    )

    model.wait()

    print(model.display_name)
    print(model.resource_name)
    return model



import urllib
import json

import google.auth.transport.requests
import google.oauth2.id_token
import uuid

from google.cloud import storage


def execute_local_notebook(gcp_project: str, 
                            location: str, 
                            input_notebook_file_path: str, 
                            gcs_notebook_folder_path: str, 
                            execution_id="", 
                            env_uri="gcr.io/deeplearning-platform-release/base-cu110:latest", 
                            kernel="python3", 
                            master_type="n1-standard-4"):
    gcs_bucket_name = _get_gcs_bucket_name_from_gcs_uri(gcs_notebook_folder_path)
    file_name = input_notebook_file_path.split("/")[-1]
    gcs_out_path = "/".join(gcs_notebook_folder_path.replace("gs://", "").split("/")[1:]) + "/" + file_name
    input_gcs_notebook_path = f"gs://{gcs_notebook_folder_path}/{file_name}"
    _upload_blob(gcp_project, gcs_bucket_name, input_notebook_file_path, gcs_out_path)
    return execute_notebook(gcp_project, location, input_gcs_notebook_path, gcs_notebook_folder_path, execution_id, env_uri, kernel, master_type)


def execute_notebook(gcp_project: str, 
                     location: str, 
                     gcs_input_notebook_file_path: str, 
                     gcs_output_notebook_folder_path: str, 
                     execution_id="", 
                     env_uri="gcr.io/deeplearning-platform-release/base-cu110:latest", 
                     kernel="python3", 
                     master_type="n1-standard-4"):
    if not execution_id:
        execution_id = str(uuid.uuid1())

    service_url = f"https://notebooks.googleapis.com/v1/projects/{gcp_project}/locations/{location}/executions?execution_id={execution_id}"
    values = {
        "description": f"Execution for {gcs_input_notebook_file_path}",
        "executionTemplate": {
            "scaleTier": "CUSTOM",
            "masterType": master_type,
            "inputNotebookFile": gcs_input_notebook_file_path,
            "outputNotebookFolder": gcs_output_notebook_folder_path,
            "containerImageUri": env_uri,
            "kernelSpec": kernel
        }
    }
    data = json.dumps(values).encode('utf-8')
    data_from_gcp = _send_generic_request(service_url, data)
    if not data_from_gcp:
        return None
    operation_uri = data_from_gcp["name"]
    execution_uri = data_from_gcp["metadata"]["target"]
    notebook_gcs_url = get_output_notebook_path(execution_uri)
    notebook_gcs_url_without_scheme = notebook_gcs_url.replace("gs://", "")
    viewer_url = f"https://notebooks.cloud.google.com/view/{notebook_gcs_url_without_scheme}"
    return {
        "operation_uri": operation_uri,
        "execution_uri": execution_uri,
        "notebook_gcs_url": notebook_gcs_url,
        "viewer_url": viewer_url
    }


def get_output_notebook_path(execution_uri: str) -> str:
    reuqest_url = f"https://notebooks.googleapis.com/v1/{execution_uri}"
    response = _send_generic_request(reuqest_url)
    return response["outputNotebookFile"]


def _upload_blob(gcp_project, bucket_name, source_file_name, destination_blob_name):
    storage_client = storage.Client(project=gcp_project)
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    blob.upload_from_filename(source_file_name)
    print(
        "File {} uploaded to {}.".format(
            source_file_name, destination_blob_name
        )
    )


def _get_gcs_bucket_name_from_gcs_uri(gcs_uri):
    return gcs_uri.split("/")[2]


def _send_generic_request(url, data=None):
    creds, _ = google.auth.default()
    auth_req = google.auth.transport.requests.Request()
    creds.refresh(auth_req)
    req = urllib.request.Request(url, data=data)
    req.add_header('Content-Type', 'application/json')
    req.add_header("Authorization", f"Bearer {creds.token}")
    response = urllib.request.urlopen(req)
    if response.status != 200:
        print(f"Error: {response.status}")
        print(f"Error: {response.read()}")
        return None
    encoding = response.info().get_content_charset('utf-8')
    return json.loads(response.read().decode(encoding))


# if "__main__" == __name__:
    # https://notebooks.googleapis.com/v1/projects/ml-lab-152505/locations/us-central1/executions?execution_id=3a5c9802-8f8d-11ec-b585-acde48001122
    # https://notebooks.googleapis.com/v1/projects/ml-lab-152505/locations/us-central1/executions?execution_id=0e05f924-8f8d-11ec-9223-acde48001122
    # :path: /aipn/v2/proxy/notebooks.googleapis.com%2Fv1%2Fprojects%2Fml-lab-152505%2Flocations%2Fus-central1%2Fexecutions%3Fexecution_id%3Duntitled__1645052627007?1645052645915
    # URL: https://7cc62a62987d13d7-dot-us-central1.notebooks.googleusercontent.com/aipn/v2/proxy/notebooks.googleapis.com%2Fv1%2Fprojects%2Fml-lab-152505%2Flocations%2Fus-central1%2Fexecutions%3Fexecution_id%3Duntitled__1645052627007?1645052645915
    # print(str(execute_notebook("ml-lab-152505", "us-central1", "gs://test-bucket-for-notebooks/executor_files/untitled__1645052627007/Untitled.ipynb", "gs://test-bucket-for-notebooks/executor_files/untitled__1645052627007")))
    # notebook_gcs_url = get_output_notebook_path("projects/ml-lab-152505/locations/us-central1/executions/83214dfe-90d1-11ec-bd9d-acde48001122")
    # notebook_gcs_url_without_scheme = notebook_gcs_url.replace("gs://", "")
    # viewer_url = f"https://notebooks.cloud.google.com/view/{notebook_gcs_url_without_scheme}"
    # print(viewer_url)
    # print(execute_local_notebook("ml-lab-152505", "us-central1", "/Users/vsk/src/notebooks-ci-showcase/notebooks/clean.ipynb", "gs://test-bucket-for-notebooks/executor_files/"))



from flask import Flask, request, jsonify
from prediction import predict

app = Flask(__name__)


@app.route("/predict", methods=["POST"])
def index():
    json_data = request.json
    kwarg = json_data["parameters"]
    instances = json_data["instances"]
    results = []
    for instance in instances:
        results.append(predict(instance, **kwarg))
    response = {
        "predictions": results
    }
    return jsonify(response)


@app.route("/ping")
def ping():
    return "pong"


import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 8):
    sys.exit("Sorry, Python < 3.8 is not supported")


setup(
    name="ai-workbench",
    version="1.0.0",
    author="Viacheslav Kovalevskyi",
    author_email="viacheslav@kovalevskyi.com",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "workbench=workbench.cli:main"
        ]
    },
    url="http://pypi.python.org/pypi/nmp/",
    license="LICENSE",
    description="GCP Vertex AI high level SDK",
    install_requires=[
        "docker==5.0.*",
        "google-cloud-aiplatform==1.1.*",
        "grpcio==1.36.*",
        "PyYAML==5.4.*"
    ],
    include_package_data=True,
    package_data={"workbench": ["workbench/prediction/container/Dockerfile", "workbench/prediction/container/requirements.txt"]},
)



import os
import shutil
from typing import Optional
from workbench.prediction import constants

import docker
import tempfile


def create_model(tag: str,
                 src_folder: Optional[str] = "."):
    with tempfile.TemporaryDirectory() as tmpdir:
        target_dir = os.path.join(
            tmpdir, constants.TEMP_DIR_FOR_BUILDING_CONTAINERS)
        print("Preparing Docker env")
        _check_if_prediction_file_exist(src_folder)
        print("Preparing Docker env")
        _move_docker_content_to_temp_dir(target_dir)
        print("Moving content of the current dir to the temp location")
        shutil.copytree(src_folder, target_dir, dirs_exist_ok=True)
        print("Building and pushing docker container")
        _build_and_push_docker(target_dir, tag)
        print("done")


def _check_if_prediction_file_exist(target_dir):
    prediction_file = os.path.join(target_dir, constants.MAIN_PREDICTION_FILE)
    if not os.path.isfile(prediction_file):
        raise ValueError(
            "The prediction file {} does not exist".format(prediction_file))


def _move_docker_content_to_temp_dir(target_dir):
    src = os.path.join(
        os.path.dirname(__file__), constants.PACKAGE_DIR_FOR_CONTAINER)
    shutil.copytree(src, target_dir, dirs_exist_ok=True)


def _build_and_push_docker(path, tag):
    client = docker.from_env()
    client.images.build(path=path, tag=tag)
    client.images.push(tag)


# create_model("us.gcr.io/ml-lab-152505/model-poc", main_nb_name="test.ipynb")
# create_model()
